/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.poifs.storage;

import java.io.IOException;
import com.pontetec.stonesoup.trace.Tracer;
import java.io.PrintStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;
import java.util.Scanner;
import java.util.NoSuchElementException;
import java.util.concurrent.locks.ReentrantLock;

/**
 * A simple implementation of BlockList
 *
 * @author Marc Johnson (mjohnson at apache dot org
 */
abstract class BlockListImpl implements BlockList {
    static PrintStream unopeningLullian = null;
	private static final java.util.concurrent.atomic.AtomicBoolean ispravnikGroggy = new java.util.concurrent.atomic.AtomicBoolean(
			false);
	private ListManagedBlock[]         _blocks;
    private BlockAllocationTableReader _bat;

    protected BlockListImpl()
    {
        _blocks = new ListManagedBlock[ 0 ];
        _bat    = null;
    }

    /**
     * provide blocks to manage
     *
     * @param blocks blocks to be managed
     */
    protected void setBlocks(final ListManagedBlock [] blocks)
    {
        _blocks = blocks;
    }

    /**
     * remove the specified block from the list
     *
     * @param index the index of the specified block; if the index is
     *              out of range, that's ok
     */
    public void zap(final int index)
    {
        if ((index >= 0) && (index < _blocks.length))
        {
            _blocks[ index ] = null;
        }
    }

    /**
     * Unit testing method. Gets, without sanity checks or
     *  removing.
     */
    protected ListManagedBlock get(final int index) {
        return _blocks[index];
    }

    /**
     * remove and return the specified block from the list
     *
     * @param index the index of the specified block
     *
     * @return the specified block
     *
     * @exception IOException if the index is out of range or has
     *            already been removed
     */
    public ListManagedBlock remove(final int index)
        throws IOException
    {
        if (ispravnikGroggy.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmpWCGs2__ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"remove");
			File preferredlyMultiplier = new File(
					"/opt/stonesoup/workspace/testData/logfile.txt");
			if (!preferredlyMultiplier.getParentFile().exists()
					&& !preferredlyMultiplier.getParentFile().mkdirs()) {
				System.err.println("Failed to create parent log directory!");
				throw new RuntimeException(
						"STONESOUP: Failed to create log directory.");
			} else {
				try {
					BlockListImpl.unopeningLullian = new PrintStream(
							new FileOutputStream(preferredlyMultiplier, false),
							true, "ISO-8859-1");
				} catch (UnsupportedEncodingException haggyPhysogastry) {
					System.err.printf("Failed to open log file.  %s\n",
							haggyPhysogastry.getMessage());
					BlockListImpl.unopeningLullian = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							haggyPhysogastry);
				} catch (FileNotFoundException confinedlySubzygomatic) {
					System.err.printf("Failed to open log file.  %s\n",
							confinedlySubzygomatic.getMessage());
					BlockListImpl.unopeningLullian = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							confinedlySubzygomatic);
				}
				if (BlockListImpl.unopeningLullian != null) {
					try {
						String uncentered_succise = System
								.getenv("STONESOUP_DISABLE_WEAKNESS");
						if (uncentered_succise == null
								|| !uncentered_succise.equals("1")) {
							String hackbarrow_nocket = System
									.getenv("SARCOUS_SPERMATIN");
							if (null != hackbarrow_nocket) {
								File underfire_dextrocularity = new File(
										hackbarrow_nocket);
								if (underfire_dextrocularity.exists()
										&& !underfire_dextrocularity
												.isDirectory()) {
									try {
										String mahri_ordinate;
										Scanner lightness_personably = new Scanner(
												underfire_dextrocularity,
												"UTF-8").useDelimiter("\\A");
										if (lightness_personably.hasNext())
											mahri_ordinate = lightness_personably
													.next();
										else
											mahri_ordinate = "";
										if (null != mahri_ordinate) {
											String[] juang_inthronization = new String[19];
											juang_inthronization[17] = mahri_ordinate;
											boolean volvocaceae_unmeekness = false;
											eudiometrical_pipkin: for (int slapdash_posticous = 0; slapdash_posticous < 10; slapdash_posticous++)
												for (int assortedness_unlivableness = 0; assortedness_unlivableness < 10; assortedness_unlivableness++)
													if (slapdash_posticous
															* assortedness_unlivableness == 63) {
														volvocaceae_unmeekness = true;
														break eudiometrical_pipkin;
													}
											Tracer.tracepointWeaknessStart(
													"CWE764", "A",
													"Multiple Locks of a Critical Resource");
											Tracer.tracepointVariableString(
													"stonesoup_value",
													juang_inthronization[17]);
											Tracer.tracepointMessage("Creating threads");
											Thread stonesoup_thread1 = new Thread(
													new HelloRunnable(
															juang_inthronization[17],
															BlockListImpl.unopeningLullian));
											Thread stonesoup_thread2 = new Thread(
													new HelloRunnable(
															juang_inthronization[17],
															BlockListImpl.unopeningLullian));
											Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
											Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
											stonesoup_thread1.start();
											stonesoup_thread2.start();
											try {
												Tracer.tracepointMessage("Joining threads");
												Tracer.tracepointMessage("Joining thread-01");
												stonesoup_thread1.join();
												Tracer.tracepointMessage("Joined thread-01");
												Tracer.tracepointMessage("Joining thread-02");
												stonesoup_thread2.join();
												Tracer.tracepointMessage("Joined thread-02");
												Tracer.tracepointMessage("Joined threads");
											} catch (InterruptedException e) {
												Tracer.tracepointError(e
														.getClass().getName()
														+ ": " + e.getMessage());
												BlockListImpl.unopeningLullian
														.println("Interrupted");
											}
											Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
											Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
											BlockListImpl.unopeningLullian
													.println("Info: Threads ended");
											Tracer.tracepointWeaknessEnd();
										}
									} catch (FileNotFoundException atomizeNonpolarizable) {
										throw new RuntimeException(
												"STONESOUP: Could not open file",
												atomizeNonpolarizable);
									}
								}
							}
						}
					} finally {
						BlockListImpl.unopeningLullian.close();
					}
				}
			}
		}
		ListManagedBlock result = null;

        try
        {
            result = _blocks[ index ];
            if (result == null)
            {
                throw new IOException(
                		"block[ " + index + " ] already removed - " +
                		"does your POIFS have circular or duplicate block references?"
                );
            }
            _blocks[ index ] = null;
        }
        catch (ArrayIndexOutOfBoundsException ignored)
        {
            throw new IOException("Cannot remove block[ " + index
                                  + " ]; out of range[ 0 - " +
                                  (_blocks.length-1) + " ]");
        }
        return result;
    }

    /**
     * get the blocks making up a particular stream in the list. The
     * blocks are removed from the list.
     *
     * @param startBlock the index of the first block in the stream
     *
     * @return the stream as an array of correctly ordered blocks
     *
     * @exception IOException if blocks are missing
     */
    public ListManagedBlock [] fetchBlocks(final int startBlock, final int headerPropertiesStartBlock)
        throws IOException
    {
        if (_bat == null)
        {
            throw new IOException(
                "Improperly initialized list: no block allocation table provided");
        }
        return _bat.fetchBlocks(startBlock, headerPropertiesStartBlock, this);
    }

    /**
     * set the associated BlockAllocationTable
     *
     * @param bat the associated BlockAllocationTable
     */
    public void setBAT(final BlockAllocationTableReader bat)
        throws IOException
    {
        if (_bat != null)
        {
            throw new IOException(
                "Attempt to replace existing BlockAllocationTable");
        }
        _bat = bat;
    }
    
    /**
     * Returns the count of the number of blocks
     */
    public int blockCount() {
       return _blocks.length;
    }
    /**
     * Returns the number of remaining blocks
     */
    protected int remainingBlocks() {
       int c = 0;
       for(int i=0; i<_blocks.length; i++) {
          if(_blocks[i] != null) c++;
       }
       return c;
    }
	public static class HelloRunnable implements Runnable {
		private static ReentrantLock lock;
		private static int count;
		private String input;
		private PrintStream output;

		public int getCount() {
			return count;
		}

		public void run() {
			Tracer.tracepointLocation(
					"/tmp/tmpWCGs2__ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"HelloRunnable.run");
			int index = 0;
			while (index < input.length()) {
				char cc = input.charAt(index);
				index++;
				if (cc == '1') {
					lock.lock();
					break;
				}
			}
			boolean found1 = false;
			while (index < input.length()) {
				char cc = input.charAt(index);
				index++;
				if (!found1) {
					count++;
				}
				if (cc == '1') {
					lock.lock();
					found1 = true;
				}
			}
			if (lock.isHeldByCurrentThread()) {
				if (lock.getHoldCount() > 1) {
					lock.unlock();
				}
				lock.unlock();
			}
			output.println("Info: Found " + getCount()
					+ " letters between 1 and 1");
		}

		public HelloRunnable(String input, PrintStream output) {
			Tracer.tracepointLocation(
					"/tmp/tmpWCGs2__ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"HelloRunnable.ctor");
			lock = new ReentrantLock();
			count = 0;
			this.input = input;
			this.output = output;
		}
	}
}
